import pytest
import material_mechanics as mm


@pytest.fixture(scope='module')
def ortho_mat():
    name = 'Orthotropic Material'

    stiffnesses = dict(e1=100000, e2=8000, e3=7000, g12=5000, g13=5000, g23=4000)
    poissons = dict(nu12=0.33, nu21=0.02, nu13=0.33, nu31=0.02, nu23=0.33, nu32=0.02)
    strengths = None
    density = 1.0

    return mm.orthotropic_material(name=name, stiffness=stiffnesses, poisson=poissons, strength=strengths, density=density)


@pytest.fixture(scope='module')
def gfk_mat():
    name = 'GFK'
    stiffnesses = dict(e1=45160, e2=14700, g12=5300)
    poissons = dict(nu12=0.3, nu23=0.38)
    strengths = ((1200, 900, None, None), (50, 170, 70, 70), (50, 170, 70, 70))
    density = 2.0

    return mm.transverse_isotropic_material(name=name, stiffness=stiffnesses, poisson=poissons, strength=strengths, density=density)


@pytest.fixture(scope='module')
def cfk_mat():
    name = 'CFK'
    stiffnesses = dict(e1=140000, e2=9000, g12=4600)
    poissons = dict(nu12=0.3, nu23=0.37)
    strengths = None
    density = 1.5

    material = mm.transverse_isotropic_material(name=name, stiffness=stiffnesses, poisson=poissons, strength=strengths, density=density)

    return material


@pytest.fixture(scope='module')
def fiber_eglas():
    name = 'E Glass Fiber'
    material = mm.isotropic_material(
        name=name,
        stiffness=72000.0,
        poisson=0.22,
        strength=None,
        density=2.54
    )
    return material


@pytest.fixture(scope='module')
def fiber_carbon_ht():
    name = 'Carbon Fiber HT'
    stiffness = dict(
        e1=230000,
        e2=13000,
        g12=50000
    )

    poisson = dict(
        nu12=0.23,
        nu23=0.3
    )

    material = mm.transverse_isotropic_material(
        name=name,
        stiffness=stiffness,
        poisson=poisson,
        strength=None,
        density=1.74
    )
    return material


@pytest.fixture(scope='module')
def fiber_carbon_hm():
    name = 'Carbon Fiber HM'
    stiffness = dict(
        e1=392000,
        e2=10000,
        g12=30000
    )

    poisson = dict(
        nu12=0.2,
        nu23=0.3
    )

    material = mm.transverse_isotropic_material(
        name=name,
        stiffness=stiffness,
        poisson=poisson,
        strength=None,
        density=1.81
    )
    return material


@pytest.fixture(scope='module')
def matrix_pa66():
    name = 'PA66'
    material = mm.isotropic_material(
        name=name,
        stiffness=2000.0,
        poisson=0.4,
        strength=65.0,
        density=1.13
    )
    return material


@pytest.fixture(scope='module')
def matrix_ep():
    name = 'Epoxy Resin'
    material = mm.isotropic_material(
        name=name,
        stiffness=3200.0,
        poisson=0.3,
        strength=90.0,
        density=1.2
    )
    return material


@pytest.fixture(scope='module')
def frp_cfk(fiber=fiber_carbon_ht(), matrix=matrix_ep(), phi=0.65):
    # set material
    strength_dict = dict(
        r_11_tensile=2000.0,
        r_11_compression=1650.0,
        r_22_tensile=70.,
        r_22_compression=240.,
        r_12=105,
    )

    material = mm.FiberReinforcedPlastic(fiber_material=fiber, matrix_material=matrix, fiber_volume_fraction=phi,
                                         name='CFRP', symmetry='mean', strength=strength_dict)

    return material


@pytest.fixture(scope='module')
def frp_gfk(fiber=fiber_eglas(), matrix=matrix_pa66(), phi=0.5):
    # set material
    material = mm.FiberReinforcedPlastic(fiber_material=fiber, matrix_material=matrix, fiber_volume_fraction=phi,
                                         name='GFRP', symmetry='mean')

    return material


@pytest.fixture(scope='module')
def fzb_lam(frp_cfk):
    lam_fzb = mm.Laminate()
    lam_fzb.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=45.))
    lam_fzb.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=90.))
    lam_fzb.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=135.))
    lam_fzb.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=0.))
    lam_fzb.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=0.))
    lam_fzb.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=135.))
    lam_fzb.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=90.))
    lam_fzb.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=45.))
    yield lam_fzb
    del lam_fzb


@pytest.fixture(scope='module')
def x_lam(frp_cfk):
    lam_x = mm.Laminate()
    lam_x.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=0.))
    lam_x.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=90.))
    lam_x.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=0.))
    lam_x.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=90.))
    lam_x.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=90.))
    lam_x.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=0.))
    lam_x.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=90.))
    lam_x.add_layer(mm.Layer(material=frp_cfk, thickness=0.25, orientation=0.))
    yield lam_x
    del lam_x


@pytest.fixture(scope='module')
def asym_lam(m1=frp_cfk(), m2=frp_gfk(), t=0.25):
    laminate_creator = mm.StandardLaminateFactory(material=m1, layer_thickness=t)

    layers = [
        dict(orientation=0),
        dict(orientation=30),
        dict(orientation=8, material=m2),
        dict(orientation=12),
        dict(orientation=-20, thickness=0.2),
        dict(orientation=33),
        dict(orientation=90),
        dict(orientation=44, material=m2, thickness=0.5),
    ]

    lam = laminate_creator.get_laminate(layers=layers)

    return lam


@pytest.fixture(scope='module')
def puck_set_frp_cfk(mat=frp_cfk()):
    puck_set = mm.PuckStrengthSet(material=mat)
    yield puck_set
    del puck_set



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module holds possible material laws
"""
import logging
import numpy as np

from ..tools import force_symmetry


# ====================
# Set up module logger
# ====================
logger = logging.getLogger(__name__)


class HookesLaw:
    r"""
    representation of Hookes Law of elasticity for two- and three-dimensional stress states in materials

    .. note::
        poisson ratios are in international notation.
        The first index points to the causing strain.
        The second index points to the resulting strain.

    :param stiffness: List or tuples of pairs of stiffness values. Each pair holds the stiffness in a major axis of the
        material and the shear stiffness in the plane perpendicular to that axis.
        ((:math:`\sigma_{11}`, :math:`G_{23}`) for the 1-direction)
    :param poisson: List or tuples of pairs of poisson ratios. each pair holds the two poisson ratios perpendicular to a
        major material axis. So the first entry holds (:math:`\nu_{23}`, :math:`\nu_{32}`).
    :type stiffness: List or tuple of lists or tuples of floats
    :type poisson: List or tuple of lists or tuples of floats
    """

    def __init__(self, stiffness, poisson, *args, **kwargs):
        self.symmetry = kwargs.get('symmetry', 'mean')

        self.e1 = stiffness[0][0]
        self.e2 = stiffness[1][0]
        self.e3 = stiffness[2][0]

        self.g23 = stiffness[0][1]
        self.g13 = stiffness[1][1]
        self.g12 = stiffness[2][1]

        self.nu23 = poisson[0][0]
        self.nu32 = poisson[0][1]
        self.nu13 = poisson[1][0]
        self.nu31 = poisson[1][1]
        self.nu12 = poisson[2][0]
        self.nu21 = poisson[2][1]

    @property
    def compliance_matrix(self):
        """
        returns the compliance matrix for three-dimensional stress states

        :return: compliance matrix (6x6)
        :rtype: numpy.ndarray
        """
        compliance_matrix = np.zeros((6, 6))
        compliance_matrix[0, 0] = 1. / self.e1
        compliance_matrix[0, 1] = -self.nu21 / self.e2
        compliance_matrix[0, 2] = -self.nu31 / self.e3
        compliance_matrix[1, 0] = -self.nu12 / self.e1
        compliance_matrix[1, 1] = 1. / self.e2
        compliance_matrix[1, 2] = -self.nu32 / self.e3
        compliance_matrix[2, 0] = -self.nu13 / self.e1
        compliance_matrix[2, 1] = -self.nu23 / self.e2
        compliance_matrix[2, 2] = 1. / self.e3
        compliance_matrix[3, 3] = 1. / self.g23
        compliance_matrix[4, 4] = 1. / self.g13
        compliance_matrix[5, 5] = 1. / self.g12
        return compliance_matrix

    @property
    def compliance_matrix_2d(self):
        """
        returns the compliance matrix for two-dimensional stress states

        :return: compliance matrix (3x3)
        :rtype: numpy.ndarray
        """
        compliance_matrix = np.zeros((3, 3))
        compliance_matrix[0, 0] = 1. / self.e1
        compliance_matrix[0, 1] = -self.nu21 / self.e2
        compliance_matrix[1, 0] = -self.nu12 / self.e1
        compliance_matrix[1, 1] = 1. / self.e2
        compliance_matrix[2, 2] = 1. / self.g12
        return compliance_matrix

    @property
    def stiffness_matrix(self):
        """
        returns the stiffness matrix for three-dimensional stress states

        :return: stiffness matrix (6x6)
        :rtype: numpy.ndarray
        """
        return np.linalg.inv(self.compliance_matrix)

    @property
    def stiffness_matrix_2d(self):
        """
        returns the stiffness matrix for two-dimensional stress states

        :return: stiffness matrix (3x3)
        :rtype: numpy.ndarray
        """
        return np.linalg.inv(self.compliance_matrix_2d)



"""
This module provides all necessary tools to calculate the two- and three dimensional puck material exertions of fiber
reinforced plastics. For further reading and details on the theory please check out the refereed literature.

.. rubric:: Literature

.. [Sch07] *German* H. Schürmann, Konstruieren mit Faser-Kunststoff-Verbunden. Berlin, Heidelberg, 2007.
.. [Puc04] *English* A. Puck and H. Schürmann, “Failure analysis of FRP laminates by means of physically based
    phenomenological models,” Fail. Criteria Fibre-Reinforced-Polymer Compos., pp. 832–876, Jan. 2004.
.. [Puc02] *English* [1] A. Puck, J. Kopp, and M. Knops, “Guidelines for the determination of the parameters in Puck’s
    action plane strength criterion,” Compos. Sci. Technol., vol. 62, no. 3, pp. 371–378, Feb. 2002.
"""
import logging
import numpy as np

from .. import FiberReinforcedPlastic

# ====================
# Set up module logger
# ====================
logger = logging.getLogger(__name__)


class PuckStrengthSet(object):
    r"""
    Class providing the necessary methods to calculate the Puck Criterion

    For further details please check [Sch07]_ [Puc04]_ [Puc02]_

    :param material: composite material for the puck strength criterion
    :param p_plp: (*optional*) slope parameter  :math:`p_{\perp\parallel}^+`, default: 0.27
    :param p_plm: (*optional*) slope parameter  :math:`p_{\perp\parallel}^-`, default: 0.27
    :param p_ppp: (*optional*) slope parameter  :math:`p_{\perp\perp}^+`, default: 0.3
    :param p_ppm: (*optional*) slope parameter  :math:`p_{\perp\perp}^-`, default: 0.35
    :param m_sf: (*optional*) magnification factor :math:`m_{\sigma,f}` for fiber strain :math:`\varepsilon_{1}` due to
        perpendicular stress (:math:`\sigma_{2}`, :math:`\sigma_{3}`), default: 1.1
    :type material: FiberReinforcedPlastic
    :type p_plp: float
    :type p_plm: float
    :type p_ppp: float
    :type p_ppm: float
    :type m_sf: float
    """

    def __init__(self, material, *args, **kwargs):
        assert isinstance(material, FiberReinforcedPlastic)

        # material properties
        self.nu_f_ql = material.fiber_material.get_poisson(index=12)
        self.e_l = material.get_stiffness()
        try:
            self.e_f_l = material.fiber_material.get_stiffness()
        except AttributeError:
            self.e_f_l = kwargs.get('e_fiber', None)
        self.nu_ql = material.get_poisson(index=12)
        try:
            self._phi = material.fiber_volume_fraction
        except AttributeError:
            self._phi = kwargs.get('phi', None)

        # curve parameters
        self.p_ppm = kwargs.get('p_ppm', 0.27)
        self.p_ppp = kwargs.get('p_ppp', 0.27)
        self.p_plm = kwargs.get('p_plm', 0.3)
        self.p_plp = kwargs.get('p_plp', 0.35)
        self.m_sf = kwargs.get('m_qf', 1.1)

        # Strength properties
        self.r_lp = material.get_strength(index=11, direction='+')
        self.r_lm = material.get_strength(index=11, direction='-')
        self.r_pp = material.get_strength(index=22, direction='+')
        self.r_pm = material.get_strength(index=11, direction='-')
        self.r_pl = material.get_strength(index=12)
        self.r_ppa = self.r_pm / (2 * (1 + self.p_ppm))

    def inter_fiber_exertion(self, stress_vector, theta):
        r"""
        calculates the inter-fiber exertion in the plane defined by the angle :math:`\Theta`

        :param stress_vector: stress in fiber coordinates (:math:`\sigma_{11}`, :math:`\sigma_{22}`,
            :math:`\sigma_{33}`, :math:`\sigma_{23}`, :math:`\sigma_{31}`, :math:`\sigma_{21}`)
        :param theta: angle defining the plane in radians
        :type stress_vector: numpy.ndarray
        :type theta: float
        :return: inter fiber exertion
        """

        relevant_stresses = stress_vector[1:]
        if all(relevant_stresses == 0):
            return 0

        sig_n, tau_nt, tau_n1 = np.dot(get_stress_transformation_matrix(theta), relevant_stresses)

        if tau_nt == 0:
            c = 0
        else:
            c = tau_nt ** 2 / (tau_nt ** 2 + tau_n1 ** 2)

        if tau_n1 == 0:
            s = 0
        else:
            s = tau_n1 ** 2 / (tau_nt ** 2 + tau_n1 ** 2)

        if sig_n >= 0:
            form_factor = self.p_ppp / self.r_ppa * c + self.p_plp / self.r_pl * s
            exertion = np.sqrt(((1. / self.r_pp - form_factor) * sig_n) ** 2 + (tau_nt / self.r_ppa) ** 2 + (
                tau_n1 / self.r_pl) ** 2) + form_factor * sig_n
        else:
            form_factor = self.p_ppm / self.r_ppa * c + self.p_plm / self.r_pl * s
            exertion = np.sqrt((tau_nt / self.r_ppa) ** 2 + (tau_n1 / self.r_pl) ** 2 + (
                form_factor * sig_n) ** 2) + form_factor * sig_n

        return exertion

    def get_max_inter_fiber_exertion(self, stress_vector, *args, **kwargs):
        r"""
        returns the maximum inter fiber strain and the angle of the plane in which it occurs

        :param stress_vector: stress vector in th fiber coordinate system (:math:`\sigma_{11}`, :math:`\sigma_{22}`,
            :math:`\sigma_{33}`, :math:`\sigma_{23}`, :math:`\sigma_{31}`, :math:`\sigma_{21}`)
        :param precision: (*optional*) number of decimal points to which to return the calculated inter-fiber exertion,
            default is 2
        :param angle_precision: (*optional*) number of decimal points to which to calculate the angle of the plane in
            which maximum strain occurs, default is 2
        :type stress_vector: numpy.ndarray
        :type precision: int
        :type angle_precision: int
        :return: 2-tuple

            - maximum inter fiber strain
            - angle of plane in which maximum strain occurs
        """
        assert isinstance(stress_vector, np.ndarray)

        precision = kwargs.get('precision', 2)
        angle_precision = kwargs.get('angle_precision', 2)

        relevant_stresses = stress_vector[1:]
        if all(relevant_stresses == 0):
            return 0

        offset = 0.0
        func = lambda x: 1. / (offset + self.inter_fiber_exertion(stress_vector, x * np.pi / 180.))
        inter_fiber_exertion, fracture_angle = find_min_stress_angle(func, offset=offset, precision=angle_precision)
        return round(inter_fiber_exertion, precision), fracture_angle

    def get_fiber_exertion(self, stress_vector, *args, **kwargs):
        r"""
        returns fiber strain for provided stress vector

        :param stress_vector: stress vector in fiber coordinate system (:math:`\sigma_{11}`, :math:`\sigma_{22}`,
            :math:`\sigma_{33}`, :math:`\sigma_{23}`, :math:`\sigma_{31}`, :math:`\sigma_{21}`)
        :param precision: (*optional*) number of decimal points to which to return the calculated fiber exertion,
            default is 2
        :type stress_vector: numpy.ndarray
        :type precision: int
        :return: fiber strain
        """
        assert isinstance(stress_vector, np.ndarray)

        precision = kwargs.get('precision', 2)

        if all(stress_vector[:3] == 0):
            return 0.
        sigma1 = stress_vector[0]
        sigma2 = stress_vector[1]
        sigma3 = stress_vector[2]

        if sigma1 >= 0:
            exertion = np.abs(1. / self.r_lp * (sigma1 - (self.nu_ql - self.nu_f_ql * self.e_l / self.e_f_l * self.m_sf) * (sigma2 + sigma3)))
        else:
            exertion = np.abs(1. / self.r_lm * (sigma1 - (self.nu_ql - self.nu_f_ql * self.e_l / self.e_f_l * self.m_sf) * (sigma2 + sigma3)))

        return round(exertion, precision)

    def puck_exertion_2d(self, stress_vector, *args, **kwargs):
        r"""
        calculates the puck exertions under the assumption of a two-dimensional stress state

        For details on the meaning of the parameters s and m, please check [Sch07]_ [Puc04]_ [Puc02]_

        :param stress_vector: stress vector in fiber coordinate system (:math:`\sigma_{11}`, :math:`\sigma_{22}`,
            :math:`\sigma_{33}`, :math:`\sigma_{23}`, :math:`\sigma_{31}`, :math:`\sigma_{21}`)
        :param s: (*optional*) fraction of fiber exertion at which :math:`\sigma_{11}` starts to influence damage initiation
        :param m: (*optional*) minimal value of :math:`\eta_w` at which damage due to fiber and inter-fiber exertion simultaniously
            occur.
        :param ret_type: (*optional*) sets the return format for the results ('tuple', 'array', 'dict'), default is
            'tuple'
        :param precision: (*optional*) number of decimal points to which to return the calculated exertions,
            default is 2
        :type stress_vector: numpy.ndarray
        :type s: float
        :type m: float
        :type ret_type: str
        :type precision: int
        :return: fiber exertion (*fe_fb*), inter-fiber exertion without :math:`\sigma_{11}` influence (*fe_0*),
            inter-fiber exertion including :math:`\sigma_{11}` influence (*fe_1*), fracture mode (*mode*)

            formats:

            - 'tuple' (*default*): tuple(fe_fb, fe_0, fe_1, mode)
            - 'array': tuple(np.array([fe_fb, fe_0, fe_1]), mode)
            - 'dict': dict(fe_fb=fe_fb, fe_0=fe_0, fe_1=fe_1, mode=mode)
        """
        s = kwargs.get('s', 0.5)
        m = kwargs.get('m', 0.5)
        ret_type = kwargs.get('return_type','tuple')
        precision = kwargs.get('precision', 2)

        # check which Mode is relevant
        tau_21c = self.r_pl*np.sqrt(1+2*self.p_ppm)
        mode = None

        fe_0 = 0
        fe_fb = self.get_fiber_exertion(stress_vector=stress_vector, precision=precision)

        if stress_vector[1] >= 0:
            mode = 'A'
            fe_0 = np.sqrt((1 - self.p_plp*self.r_pp/self.r_pl)**2*(stress_vector[1]/self.r_pp)**2 + (stress_vector[2]/self.r_pl)**2) + self.p_plp*stress_vector[1]/self.r_pl

        elif stress_vector[1] < 0:
            if stress_vector[2] == 0:
                mode = 'C'
                fe_0 = ((stress_vector[2] / (2 * (1 + self.p_ppm) * self.r_pl)) ** 2 +
                        (stress_vector[1] / self.r_pm) ** 2) * self.r_pm / -stress_vector[1]
            else:
                crit1 = np.abs(stress_vector[1] / stress_vector[2])
                crit2 = np.abs(self.r_ppa / tau_21c)
                if 0 <= crit1 <= crit2:
                    mode = 'B'
                    fe_0 = np.sqrt((stress_vector[2]/self.r_pl)**2 + (self.p_plm/self.r_pl*stress_vector[1])**2) + self.p_plm/self.r_pl*stress_vector[1]
                elif 0 <= 1/crit1 <= 1/crit2:
                    mode = 'C'
                    fe_0 = ((stress_vector[2]/(2*(1 + self.p_ppm)*self.r_pl))**2 + (stress_vector[1]/self.r_pm)**2)*self.r_pm/-stress_vector[1]

        if fe_fb == 0 or fe_0 == 0:
            fe_1 = fe_0

        elif fe_fb/fe_0 < s or fe_0/fe_fb < m:
            fe_1 = fe_0
        else:
            a = (1 - s) / np.sqrt(1 - m ** 2)
            c = fe_0/fe_fb
            eta_w = (c*(a*np.sqrt(c**2*(a**2 - s**2) + 1) +s)/((c*a)**2 + 1))
            fe_1 = fe_0/eta_w

        if precision is not None:
            fe_0 = round(fe_0, precision)
            fe_1 = round(fe_1, precision)

        if ret_type == 'tuple':
            return fe_fb, round(fe_0, 2), round(fe_1, 2), mode
        elif ret_type == 'dict':
            return dict(fe_fb=fe_fb, fe_0=round(fe_0, 2), fe_1=round(fe_1, 2), mode=mode)
        elif ret_type == 'array':
            return np.array([fe_fb, round(fe_0, 2), round(fe_1, 2)]), mode
        else:
            return fe_fb, round(fe_0, 2), round(fe_1, 2), mode


def get_laminate_exertions(laminate, line_loads, *args, **kwargs):
    r"""
    calculates the material exertion of every layer of the laminate from the laminate loads.

    The provided laminate loads (line_loads) are presumed to be line loads in :math:`\frac{N}{mm}` and line moments in
    :math:`N`, defined in the laminate coordinate system. The function returns maximum values for the fiber and
    inter-fiber exertions of all layers as well as a list of dicts, holding details for every layer.
    Every dictionary in the results list has the following entries:

        - fb: fiber-exertion
        - zfb_0: inter_fiber exertion without the influence of fiber parallel stress
        - zfb_1: inter_fiber exertion including the influence of fiber parallel stress
        - mode: the fracture mode as defined by Puck
        - stress: The stress in the layer in layer coordinates
        - strain: The strain in the layer in layer coordinates

    :param laminate: a laminate object
    :param line_loads: an array of line loads
    :type laminate: Laminate
    :type line_loads: numpy.ndarray
    :return: a tuple consisting of

        - maximum fiber exertion in any layer in the laminate
        - maximum inter-fiber exertion in any layer in the laminate
        - a list of dicts holding the results of every layer
    :rtype: tuple
    """
    result = list()
    laminate_strains = laminate.get_strains(line_loads=line_loads)
    layer_strains = laminate.get_layer_strains(laminate_strains=laminate_strains)
    max_fb = 0
    max_zfb = 0

    for layer, (layer_strain_top, layer_strain_bottom) in zip(laminate.stacking, layer_strains):
        layer_dict = dict()
        layer_puck_set = PuckStrengthSet(material=layer.material)
        layer_dict['puck'] = layer_puck_set

        # Bottom of layer
        s = np.dot(layer.material.stiffness_matrix_2d, layer_strain_bottom)
        layer_stress_bottom = np.squeeze(np.asarray(s))
        fe_fb_bottom, fe_0_bottom, fe_1_bottom, mode_bottom = layer_puck_set.puck_exertion_2d(stress_vector=layer_stress_bottom)
        layer_dict['bottom'] = dict(
            fb=fe_fb_bottom, zfb_0=fe_0_bottom, zfb_1=fe_1_bottom, mode=mode_bottom, stress=layer_stress_bottom,
            strain=layer_strain_bottom
        )
        if fe_fb_bottom > max_fb:
            max_fb = fe_fb_bottom

        if fe_1_bottom > max_zfb:
            max_zfb = fe_1_bottom

        # Top of layer
        s = np.dot(layer.material.stiffness_matrix_2d, layer_strain_top)
        layer_stress_top = np.squeeze(np.asarray(s))
        fe_fb_top, fe_0_top, fe_1_top, mode_top = layer_puck_set.puck_exertion_2d(stress_vector=layer_stress_top)
        layer_dict['top'] = dict(
            fb=fe_fb_top, zfb_0=fe_0_top, zfb_1=fe_1_top, mode=mode_top, stress=layer_stress_top,
            strain=layer_strain_top
        )
        if fe_fb_top > max_fb:
            max_fb = fe_fb_top

        if fe_1_top > max_zfb:
            max_zfb = fe_1_top

        result.append(layer_dict)
    return max_fb, max_zfb, result


def find_min_stress_angle(func, *args, **kwargs):
    """
    Find the minimal value of the provided function

    :param func: function to be minimized
    :param precision: (optional) sets the number of decimal points to which the angle is to be calculated, default is 0
    :param offset: (optional) sets the offset used in the provided func, default is 1.0
    :type func: function
    :type precision: int
    :type offset: float
    :return:
    """

    precision = kwargs.get('precision', 0)
    offset = kwargs.get('offset', 1.0)

    thetas = np.linspace(-90, 90, 19)
    values = [func(theta) for theta in thetas]

    p = 0
    delta = 10**-p
    while p <= precision:
        minimum_angel = thetas[np.argmin(values)]
        thetas = np.linspace(minimum_angel - 10 * delta, minimum_angel + 10 * delta, 21)
        values = [func(theta) for theta in thetas]
        p += 1
        delta = 10 ** -p

    return 1. / min(values) - offset, round(thetas[np.argmin(values)], precision)


def get_stress_transformation_matrix(theta, theta_in_deg=False):
    """
    returns a stress transformation matrix that transforms a stress vector from fiber coordinates into a strength plane
    with a normal orientation perpendicular to the fiber orientation. Theta defines the angle to the perpendicular orientation
    in the laminate plane.

    :param theta: rotation angle
    :param theta_in_deg: flag for parameter theta. If true unit of theta is assumed to be degree if not radians.
    :type theta: float
    :type theta_in_deg: bool
    :return: transformation matrix
    :rtype: numpy.ndarray
    """

    if theta_in_deg:
        theta = np.deg2rad(theta)

    trans_matrix = np.zeros((3, 5), dtype=float)
    c = np.cos(theta)
    s = np.sin(theta)

    trans_matrix[0, 0] = c ** 2
    trans_matrix[0, 1] = s ** 2
    trans_matrix[0, 2] = 2 * c * s

    trans_matrix[1, 0] = -c * s
    trans_matrix[1, 1] = c * s
    trans_matrix[1, 2] = c ** 2 - s ** 2

    trans_matrix[2, 3] = s
    trans_matrix[2, 4] = c

    return trans_matrix



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import copy
import abc
import logging

from .elastic_materials import ElasticMaterial
from .composites import Laminate, FiberReinforcedPlastic, Layer

# ====================
# Set up module logger
# ====================
logger = logging.getLogger(__name__)


class LayerFactory(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_layer(self, *args, **kwargs): pass


class LaminateFactory(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_laminate(self, *args, **kwargs): pass


class FrpFactory(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_material(self, *args, **kwargs): pass


class MaterialFactory(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_material(self, stiffness, poisson, strength, density, *args, **kwargs): pass


class PuckSetFactory(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_material(self, *args, **kwargs): pass


# Todo add Examples for each class/method

# ===============
# Layer Factories
# ===============
class StandardLayerFactory(LayerFactory):
    """
    Standard Factory for Layers

    Standard values for the parameters may be provided at initiation. Parameters that did not receive a value at
    initiation, need to be set when the get_layer method is called. If not a ValueError is raised.

    :param material: Layer material
    :param thickness: Thickness of the layer in :math:`mm`
    :param orientation: Orientation of the Layerin :math:`degree`
    :type material: ElasticMaterial or None
    :type thickness: float or None
    :type orientation: float or None

    .. note::

        All materials derived from
        :class:`ElasticMaterial <material_mechanics.materials.elastic_materials.ElasticMaterial>`, like
        :class:`FiberReinforcedPlastic <material_mechanics.materials.composites.FiberReinforcedPlastic>`,
        can be used in a Layer.

    Example::

        import material_mechanics as mm

        stiffness = dict(e1=140000, e2=9000, g12=4600)
        poisson = dict(nu12=0.3, nu23=0.37)
        density = 1.5

        cfk = mm.transverse_isotropic_material(
            name='CFK',
            stiffness=stiffness,
            poisson=poisson,
            strength=None,
            density=density)

        steel = mm.isotropic_material(
            name='Steel',
            stiffness=200000.0,
            poisson=0.34,
            strength=700,
            density=7.9
        )

        FixedMaterialLayerCreator = mm.StandardLayerFactory(material=cfk)
        layer_1 = FixedMaterialLayerCreator.get_layer(thickness=0.1, orientation=30)
        layer_2 = FixedMaterialLayerCreator.get_layer(thickness=0.1, orientation=60)
        layer_3 = FixedMaterialLayerCreator.get_layer(thickness=0.1, orientation=90)


        FixedThicknessLayerCreator = mm.StandardLayerFactory(thickness=0.2)
        layer_1 = FixedThicknessLayerCreator.get_layer(material=steel, orientation=30)
        layer_2 = FixedThicknessLayerCreator.get_layer(material=cfk, orientation=45)
        layer_3 = FixedThicknessLayerCreator.get_layer(material=steel, orientation=60)

        # standard values may also be overwritten
        layer_3 = FixedThicknessLayerCreator.get_layer(thickness=0.1, material=steel, orientation=60)

    """

    def __init__(self, *args, **kwargs):
        self._material = kwargs.get('material', None)
        self._thickness = kwargs.get('thickness', None)
        self._orientation = kwargs.get('orientation', None)

    def get_layer(self, *args, **kwargs):
        """
        Return a layer with a material, thickness and orientation.

        If the values for material, thickness and/or orientation where provided at initiation,
        they may be omitted or be overwritten.

        :param material: (*optional*) Layer material, default is material provided at initialization of class
        :param thickness: (*optional*) Thickness of the layer, default is thickness provided at initialization of class
        :param orientation: (*optional*) Orientation of the Layer, default is orientation provided at initialization of class
        :type material: ElasticMaterial
        :type thickness: float
        :type orientation: float
        :return: Layer
        :rtype: Layer
        """
        material = kwargs.get('material', self._material)
        thickness = kwargs.get('thickness', self._thickness)
        orientation = kwargs.get('orientation', self._orientation)

        if any([item is None for item in (material, thickness, orientation)]):
            raise ValueError('Please provide a material, thickness and orientation for the layer')

        return Layer(material=material, thickness=thickness, orientation=orientation)


# ==================
# Laminate Factories
# ==================
class StandardLaminateFactory(LaminateFactory):
    """
    Factory class for creating laminates

    :param material:
        (*optional*) Default material for each layer of the laminate.
        May be overwritten in layer dicts. Default is None

    :param layer_thickness:
        (*optional*) Default layer thickness for the laminate.
        May be overwritten in layer dicts. Default is None

    :param symmetry:
        (*optional*) symmetry mode of laminate, default is None

        Options:

        - None: No symmetry is forced
        - 'center_layer': The symmetry plane of the last layer is the symmetry plane for the laminate, so all but the last layer are mirrored by this plane
        - 'full': (*catches all strings but 'center_layer'*) all layers are mirrored at the plane defining the lower border of the last plane, including the last layer

    :type material: ElasticMaterial or derived
    :type layer_thickness: float or None
    :type symmetry: str or None
    """

    def __init__(self, *args, **kwargs):
        self._material = kwargs.get('material', None)
        self._layer_thickness = kwargs.get('layer_thickness', None)
        self._symmetry = kwargs.get('symmetry', None)
        self._layer_factory = StandardLayerFactory(material=self._material, thickness=self._layer_thickness)

    def get_laminate(self, layers, *args, **kwargs):
        r"""
        Return a Laminate

        :param layers: stacking of a laminate defined as dicts. each dict holds the parameters of the layer
            ('thickness' in :math:`mm`, 'orientation' in :math:`degree`, 'material'). 'thickness' and 'material' may be
            set at the initiation of the class, in which case they may be omitted in the layer dicts. If provided at
            method call, init values of parameters will be overwritten.
        :param symmetry: (*optional*) symmetry mode of laminate, default is None

            Options:

            - None: No symmetry is forced
            - 'center_layer': The symmetry plane of the last layer is the symmetry plane for the laminate, so all but the last layer are mirrored by this plane
            - 'full': (*catches all strings but 'center_layer'*) all layers are mirrored at the plane defining the lower border of the last plane, including the last layer

        :type layers: list of dicts
        :type symmetry: str or None
        :return: Laminate
        :rtype: Laminate
        """
        symmetry = kwargs.get('symmetry', None)
        laminate = Laminate()

        for l in layers:
            layer = self._layer_factory.get_layer(**l)
            laminate.add_layer(layer=layer)

        if symmetry is not None:
            # define symmetry layers according to provided symmetry switch
            sym_layers = list(reversed(layers))
            if symmetry == 'center_layer':
                sym_layers.pop()

            # add layers to achieve symmetry
            for layer in sym_layers:
                laminate.add_layer(layer=layer)

        return laminate


# class SymmetricalLaminateFactory(LaminateFactory):
#     """
#     Returns a symmetrical laminate
#     """
#
#     def get_laminate(self, layers, *args, **kwargs):
#         """
#         Returns a symmetric laminate
#
#         Achieves symmetry by adding the provided layers in reverse order to the layer stack.
#         The middle layer can be left out when adding the new layers to move the plain of symmetry into the middle layer.
#
#         :param layers: List of Layer objects
#         :param symmetry: flag to define position of symmetry plane
#
#             Options:
#
#                 - 0: between the two central layers (default)
#                 - 1: in the central layer (central layer is not mirrored)
#
#         :type layers: list of Layers
#         :type symmetry: (*optional*) int or None
#         :return: symmetrical laminate
#         :rtype: Laminate
#         """
#         symmetry = kwargs.get('symmetry', 0)
#         laminate = Laminate()
#
#         for layer in layers:
#             laminate.add_layer(layer=layer)
#
#         # define symmetry layers according to provided symmetry switch
#         sym_layers = list(reversed(layers))
#         if symmetry == 1:
#             sym_layers.pop()
#
#         # add layers to achieve symmetry
#         for layer in sym_layers:
#             laminate.add_layer(layer=layer)
#
#         return laminate
#
#
# class SingleMaterialLaminateFactory(LaminateFactory):
#     """
#     Factory class for creating laminates where each layer has the same material
#
#     :param material: The material for each layer of the laminate
#     :type material: ElasticMaterial
#
#     """
#
#     def __init__(self, material, *args, **kwargs):
#         self.material = material
#         self._layer_factory = StandardLayerFactory(material=material)
#
#     def get_laminate(self, layers, *args, **kwargs):
#         r"""
#         Returns a Laminate where each layer consists of the same material
#
#         :param layers: stacking of laminate defined as pairs of thickness in :math:`mm` and orientation in :math:`°`
#         :param symmetry: (*optional*) symmetry mode of laminate, default is None
#
#             options
#
#                 - None: No symmetry is forced
#                 - 'center_layer': The symmetry plane of the last layer is the symmetry plane for the laminate, so all but the last layer are mirrored by this plane
#                 - 'full': (*catches all strings but 'center_layer'*) all layers are mirrored at the plane defining the lower border of the last plane, including the last layer
#
#         :return: a Laminate with each layer having the same base material
#         :rtype: Laminate
#         """
#         symmetry = kwargs.get('symmetry', None)
#         laminate = Laminate()
#
#         for t, ori in layers:
#             layer = self._layer_factory.get_layer(material=self.material, thickness=t, orientation=ori)
#             laminate.add_layer(layer=layer)
#
#         if symmetry is not None:
#             # define symmetry layers according to provided symmetry switch
#             sym_layers = list(reversed(layers))
#             if symmetry == 'center_layer':
#                 sym_layers.pop()
#
#             # add layers to achieve symmetry
#             for layer in sym_layers:
#                 laminate.add_layer(layer=layer)
#
#         return laminate


# ==================================
# Fiber reinforced plastic factories
# ==================================
class ChangedFvcFrp(FrpFactory):
    """
    A factory class to get an FRP material from an existing FRP with changed fiber volume content

    :param material: the base FRP Material from which to calculate the generated materials
    :type material: FiberReinforcedPlastic
    """

    def __init__(self, material, *args, **kwargs):
        self._material = material

    def get_material(self, phi, *args, **kwargs):
        """
        return a fiber reinforced plastic with the same constituents as the base fiber reinforced plastic but changed
        fiber volume content

        :param phi: fiber volume content of the new fiber reinforced plastic
        :type phi: float
        :return: a fiber reinforced plastic material with the provided fiber volume content
        :rtype: FiberReinforcedPlastic
        """

        if phi is None or self._material.fiber_volume_fraction == phi:
            return self._material

        material_name = '{fiber}_{matrix}_{fvc}'.format(
            fiber=self._material.fiber_material.name,
            matrix=self._material.matrix_material.name,
            fvc=int(phi * 100)
        ).replace(" ", "")

        material = FiberReinforcedPlastic(
            name=material_name,
            fiber_material=self._material.fiber_material,
            matrix_material=self._material.matrix_material,
            fiber_volume_fraction=phi
        )

        return material


# =========
# Puck Sets
# =========
class ChangedFvcPuckSet(PuckSetFactory):
    """
    A factory class to generate puck sets from an existing puck set with changed fiber volume content

    :param puck_set: the base puck set from which to generate new puck sets
    :type puck_set: PuckStrengthSet
    """

    def __init__(self, puck_set, resin_strength):
        self._puck_set = puck_set
        self._resin_strength = resin_strength

    def get_material(self, phi, *args, **kwargs):
        """
        returns a puck set with changed fiber volume content based on the puck set provided at initiation

        :param phi: fiber volume content of the new puck set
        :type phi: float
        :return: puck set with the provided fiber volume content
        :rtype: PuckStrengthSet
        """

        if phi is None or phi == self._puck_set.phi:
            return self._puck_set

        puck_set = copy.deepcopy(self._puck_set)
        puck_set.r_lp = (self._puck_set.r_lp - self._resin_strength) * phi / self._puck_set.phi + self._resin_strength
        puck_set.r_lm = (self._puck_set.r_lm - self._resin_strength) * phi / self._puck_set.phi + self._resin_strength
        puck_set.phi = phi

        return puck_set


# =========
# Materials
# =========
def isotropic_material(stiffness, poisson, density, *args, **kwargs):
    r"""
    create a material with isotropic properties

    :param density: material density in :math:`\frac{g}{cm^3}`
    :type density: float
    :param stiffness: material stiffness in :math:`\frac{N}{mm^2}`
    :type stiffness: float
    :param poisson: material poisson's ratio
    :type poisson: float
    :param strength: material strength in :math:`\frac{N}{mm^2}`
    :type strength: float

    :return: material with isotropic properties
    :rtype: ElasticMaterial
    """
    name = kwargs.get('name', None)
    strength = kwargs.get('strength', None)

    g = stiffness / (2 * (1 + poisson))
    stiffness = [(stiffness, g)] * 3

    poisson = [(poisson, poisson)] * 3

    if strength is not None:
        strength = [strength] * 3

    material = ElasticMaterial(
        name=name,
        stiffness=stiffness,
        poisson=poisson,
        strength=strength,
        density=density
    )

    return material


def transverse_isotropic_material(stiffness, poisson, density, *args, **kwargs):
    r"""
    create a material with transverse isotropic properties

    .. note::
        poisson ratios are in international notation.
        The first index points to the causing strain.
        The second index points to the resulting strain.

    :param density: material density in :math:`\frac{g}{cm^3}`
    :type density: float
    :param stiffness: dictionary of material stiffnesses in :math:`\frac{N}{mm^2}`
    :type stiffness: dict
    :param poisson: material poisson ratios
    :type poisson: dict
    :param strength: dictionary of material strengths in :math:`\frac{N}{mm^2}`
    :type strength: dict
    :return: material with isotropic properties
    :rtype: ElasticMaterial
    """

    name = kwargs.get('name', None)
    strength = kwargs.get('strength', None)

    e1 = float(stiffness['e1'])
    e2 = float(stiffness['e2'])
    g12 = float(stiffness['g12'])

    nu12 = float(poisson['nu12'])
    nu23 = float(poisson['nu23'])
    nu21 = e2 / e1 * nu12

    g23 = e2 / (2 * (1 + nu23))

    stiffnesses = [(e1, g23), (e2, g12), (e2, g12)]
    poissons = [(nu23, nu23), (nu12, nu21), (nu12, nu21)]

    # TODO: implement strength calculation from input dictionary
    # Input type for strength is {'r1_plus': xxx, 'r1_minus': xxx, 'r2_plus': xxx, 'r2_minus': xxx, r12: xxx}

    material = ElasticMaterial(
        name=name,
        stiffness=stiffnesses,
        poisson=poissons,
        strength=strength,
        density=density)

    return material


def orthotropic_material(stiffness, poisson, density, *args, **kwargs):
    r"""
    create a material with orthotropic properties

    .. note::
        poisson ratios are in international notation.
        The first index points to the causing strain.
        The second index points to the resulting strain.

    :param density: material density in :math:`\frac{g}{cm^3}`
    :type density: float
    :param stiffness: dictionary of material stiffness in :math:`\frac{N}{mm^2}`
    :type stiffness: dict
    :param poisson: material poisson ratios
    :type poisson: dict
    :param strength: dictionary of material strengths in :math:`\frac{N}{mm^2}`
    :type strength: dict
    :return: material with isotropic properties
    :rtype: ElasticMaterial
    """

    name = kwargs.get('name', None)
    strength = kwargs.get('strength', None)

    e1 = float(stiffness['e1'])
    e2 = float(stiffness['e2'])
    e3 = float(stiffness['e3'])
    g12 = float(stiffness['g12'])
    g13 = float(stiffness['g13'])
    g23 = float(stiffness['g23'])

    nu12 = float(poisson['nu12'])
    nu21 = float(poisson['nu21'])
    nu13 = float(poisson['nu13'])
    nu31 = float(poisson['nu31'])
    nu23 = float(poisson['nu23'])
    nu32 = float(poisson['nu32'])

    # TODO: implement strength calculation from input dictionary
    # Input type for strength is {'r1_plus': xxx, 'r1_minus': xxx, 'r2_plus': xxx, 'r2_minus': xxx, r12: xxx}

    material = ElasticMaterial(
        name=name,
        stiffness=[(e1, g23), (e2, g13), (e3, g12)],
        poisson=[(nu23, nu32), (nu13, nu31), (nu12, nu21)],
        strength=strength,
        density=density)

    return material



import logging
import numpy as np

# ====================
# Set up module logger
# ====================
logger = logging.getLogger(__name__)


def get_stress_transformation_matrix(theta):
    """
    returns a stress transformation matrix that transforms a stress vector from fiber coordinates into a strength plane
    with a normal orientation perpendicular to the fiber orientation. Theta defines the angle to the perpendicular orientation
    in the laminate plane.

    :param theta: rotation angle
    :return: transformation matrix (numpy.ndarray)
    """
    trans_matrix = np.zeros((3, 5), dtype=float)
    c = np.cos(theta)
    s = np.sin(theta)

    trans_matrix[0, 0] = c ** 2
    trans_matrix[0, 1] = s ** 2
    trans_matrix[0, 2] = 2 * c * s

    trans_matrix[1, 0] = -c * s
    trans_matrix[1, 1] = c * s
    trans_matrix[1, 2] = c ** 2 - s ** 2

    trans_matrix[2, 3] = s
    trans_matrix[2, 4] = c

    return trans_matrix


def get_T_strain_2d(theta):
    """
    returns a 2d strain transformation matrix that transforms a global strain vector to a strain vector in
    fiber coordinates with a normal orientation perpendicular to the fiber orientation.
    Theta defines the angle to the perpendicular orientation in the laminate plane.

    :param theta: rotation angle
    :return: transformation matrix (numpy.ndarray)
    """
    trans_matrix = np.eye(3, 3, dtype=float)
    c = np.cos(theta)
    c2 = np.cos(2 * theta)
    s = np.sin(theta)
    s2 = np.sin(2 * theta)

    trans_matrix[0, 0] = c ** 2
    trans_matrix[0, 1] = s ** 2
    trans_matrix[0, 2] = 0.5 * s2

    trans_matrix[1, 0] = s ** 2
    trans_matrix[1, 1] = c ** 2
    trans_matrix[1, 2] = -0.5 * s2

    trans_matrix[2, 0] = -s2
    trans_matrix[2, 1] = s2
    trans_matrix[2, 2] = c2

    return trans_matrix


def get_T_strain_3d(theta):
    """
    returns a 3d strain transformation matrix that transforms a global strain vector to a strain vector in
    fiber coordinates with a normal orientation perpendicular to the fiber orientation.
    Theta defines the angle to the perpendicular orientation in the laminate plane.

    :param theta: rotation angle
    :return: transformation matrix (numpy.ndarray)
    """
    trans_matrix = np.eye(6, 6, dtype=float)
    c = np.cos(theta)
    c2 = np.cos(2 * theta)
    s = np.sin(theta)
    s2 = np.sin(2 * theta)

    trans_matrix[0, 0] = c ** 2
    trans_matrix[0, 1] = s ** 2
    trans_matrix[0, 6] = 0.5 * s2

    trans_matrix[1, 0] = s ** 2
    trans_matrix[1, 1] = c ** 2
    trans_matrix[1, 6] = -0.5 * s2

    trans_matrix[6, 0] = -s2
    trans_matrix[6, 1] = s2
    trans_matrix[6, 6] = c2

    return trans_matrix


def get_strain_at_z(global_strains, z):
    """
    returns the local strain interpolated from two given stresses

    :param global_strains:
    :param z: location at which stress should be calculated

    :return: Stress at z location
    """
    eps_0 = np.array(global_strains[:3])
    kappa_0 = np.array(global_strains[3:])

    return eps_0 + z * kappa_0


def force_symmetry(matrix, symmetry):
    r"""
    Enforce symmetry in a given matrix

    :param matrix: matrix with equal number of rows and columns
    :param symmetry: method of symmetry enforcement.

        Options:

        - None: No symmetry is being enforced
        - 'upper': upper-right elements are mirrored to lower-left elements (:math:`n_{ij} = n_{ji}; \; if: i>j`)
        - 'upper': lower-left elements are mirrored to upper-right elements (:math:`n_{ij} = n_{ji}; \; if: i<j`)
        - 'upper': upper-right elements are mirrored to lower-left elements (:math:`n_{ij} = \frac{n_{ji}+n_{ij}}{2}; \; if: i \neq j`)

    :type matrix: numpy.ndarray
    :type symmetry: str or None
    :return:
    """
    symmetric_matrix = matrix.copy()

    if symmetry is None:
        return symmetric_matrix

    for index, x in np.ndenumerate(matrix):

        if symmetry == 'upper':
            if index[0] > index[1]:
                symmetric_matrix[index] = matrix[tuple(reversed(index))]

        if symmetry == 'lower':
            if index[0] < index[1]:
                symmetric_matrix[index] = matrix[tuple(reversed(index))]

        if symmetry == 'mean':
            if index[0] != index[1]:
                symmetric_matrix[index] = np.mean((matrix[index], matrix[tuple(reversed(index))]))

    return symmetric_matrix
